=begin 

Copyright 2012 Tim Warner

Description: A Sketchup plug-in that creates the specification for a lamella roof covering a rectangular space. 

Inputs:      The required span of the roof, and its height above the edges. 
             The number of lamellas in each arch.
             The angle each arch makes with a transverse section (so-called "skew angle").
             The cross-section of the lamella, excluding the curved portion.
             Specifications of the bolt connection for the lamellas.

Outputs:     A Sketchup model of the roof
             Text output (written to the Ruby console):
                The geometry of the lamella, including the position and orientation of bolts, and cutting / drilling information.
                The geometry of the supporting beams, at the edges of the roof.
                The geometry of the braces between the lamellas at the edges of the roof.
                The geometry of the gable

Reference:   Tim Warner, "The Geometry of the Lamella Roof" (available from lamella@timwarner.ca)

Limitations: There is no error-checking on inputs. For roofs higher than a semicircle in cross-section the geometry of beams and braces
             is incorrect.
             The gable ends of the roof will be specific to the space: that is, the length of the gable lamellas is not determined here.
             The model does not include decking, insulation etc., that would make a complete roof.
             The model is a drawing tool only -- it knows nothing about engineering and structural soundness.
             
Support:     Email lamella@timwarner.ca

          
Release Notes:

30 May 2012  Release 0.1
  -------------------------------------------------------------------------
  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  -------------------------------------------------------------------------
=end

def lamella_roof

    require 'sketchup.rb'
    model=Sketchup.active_model
    ents = model.entities
    defns = model.definitions
    opts = model.rendering_options
    opts["DrawLineEnds"] = false
    selection = model.selection
    ents.clear!                       # Get rid of Sang

#  Get roof specifications

    prompts = ["Height of roof", "Span of roof", "Skew angle","Lamellas","Arches"]

    values = [1200,6000,20.0,6,2]  
    results = UI.inputbox(prompts, values, "tanzimat qaghf")

    puts "Roof Settings"
    print "Height ", results[0],"\n"
    print "Span ", results[1],"\n"
    print "Skew ", results[2],"\n"
    print "Lamellas ", results[3],"\n"
    print "Arches ", results[4],"\n"

    height = results[0].to_f
    span = results[1].to_f
    skew_angle = results[2].to_f.degrees
    lamellas = results[3].to_i
    number_of_arches = results[4].to_i
    roof_radius = (4*height**2 + span**2) / (8*height)

    # If the section higher than a semi-circle the formula for theta will be different

    if height <= span/2.0 
      theta = Math.asin(span / (2*roof_radius))/lamellas
    else
      theta = (Math::PI - Math.asin(span / (2*roof_radius)))/lamellas
    end

    initial_lamella_length = 2.0 * roof_radius * Math.sin(theta) / Math.cos(skew_angle)

#  Get lamella specifications

    prompts = ["Thickness", "Width","Bolts", "Bolt diameter"]
    values = [38.0,150.0,2,10.0]
    results = UI.inputbox(prompts, values, "andaze choobha")
    thickness = results[0].to_f
    width = results[1].to_f
    number_of_bolts = results[2].to_i
    bolt_diameter = results[3].to_f

    puts
    puts "tanzimat choobha"
    print "kolofti  ", thickness, "\n"
    print "arz ", width, "\n"
    print "Bolts ", number_of_bolts, "\n"
    print "abaad Bolt  ", bolt_diameter, "\n"

    adjusted_lamella_length = initial_lamella_length + 
                                           thickness * (1 + 1 / Math.cos(2*skew_angle)) * Math.tan(skew_angle) + 
                                           bolt_diameter*Math.tan(2*skew_angle) * Math.tan(skew_angle) + bolt_diameter

  # Compute hba = horizontal bevel angle and vba = vertical bevel angle

    hba_num = Math.sin(skew_angle)*Math.cos(skew_angle)*(1 + Math.cos(theta))
    hba_denom = 1 - (1 + Math.cos(theta)) * (Math.sin(skew_angle))**2  
    tan_hba = hba_num / hba_denom
    hba = Math.atan(tan_hba)

    vba_num = Math.tan(theta)
    vba_denom = Math.cos(skew_angle) * (1 + 1 / Math.cos(theta))
    tan_vba = vba_num / vba_denom
    vba = Math.atan(tan_vba)

    adjusted_theta = Math.asin(adjusted_lamella_length / (2*roof_radius))
    delta_width = roof_radius * (1 - Math.cos(adjusted_theta))

    lamella_blank_length = adjusted_lamella_length + 2 * thickness / tan_hba + 2 * (delta_width + width/2) * tan_vba

  # Switch dimensions to inches to draw lamella (craziness of Sketchup API)

    thickness = thickness.mm
    width = width.mm
    top_height = width/2 + delta_width.mm # figure out how much wider the lamella has to be to accommodate curvature
    length = adjusted_lamella_length.mm
    span = span.mm
    roof_radius = roof_radius.mm
    bolt_diameter = bolt_diameter.mm

    hl = length / 2.0
    ht = thickness / 2.0
    hw = width / 2.0


#  Make base lamella component

  # Create the center of the lamella, push/pull to width of lamella

    pts = []
    pts[0] = [-(hl + thickness / tan_hba),-ht, 0] 
    pts[1] = [-hl, ht, 0]
    pts[2] = [(hl + thickness / tan_hba),ht, 0]
    pts[3] = [hl,-ht, 0]

    lamella = ents.add_group()

    base  = lamella.entities.add_face(pts)
    base.material = "BurlyWood"

  # Make the top

    base.pushpull(-top_height,true)

  # Get the edges of the top face

    top = lamella.entities.select{|e| e.is_a?(Sketchup::Face) && e.plane[3] == -top_height }[0]

    top_left_edge = top.edges.select{|e| e.start.position.x < 0 && e.end.position.x < 0}[0]
    top_left_corner = top_left_edge.start

    top_right_edge = top.edges.select{|e| e.start.position.x > 0 && e.end.position.x > 0}[0]

  # Adjust top for vertical bevel

    delta = top_height * tan_vba
    movetop = Geom::Vector3d.new(-delta,0,0)
    trans = Geom::Transformation.new(movetop)
    lamella.entities.transform_entities(trans,top_left_edge)
    movetop = Geom::Vector3d.new(delta,0,0)
    trans = Geom::Transformation.new(movetop)
    lamella.entities.transform_entities(trans,top_right_edge)
  
  # Make the bottom

    base.pushpull(-hw,false)

  # Get the edges of the bottom face

    bottom = lamella.entities.select{|e| e.is_a?(Sketchup::Edge) && e.start.position.z == -hw }
    bottom_left_edge = bottom.select{|e| e.start.position.x < 0 && e.end.position.x < 0}[0]
    bottom_right_edge = bottom.select{|e| e.start.position.x > 0 && e.end.position.x > 0}[0]

  # Adjust bottom for vertical bevel

    delta = hw * tan_vba
    movebottom = Geom::Vector3d.new(delta,0,0)
    trans = Geom::Transformation.new(movebottom)
    lamella.entities.transform_entities(trans,bottom_left_edge)
    bottom_left_corner = bottom_left_edge.start
    slanted_edge = lamella.entities.add_line(bottom_left_corner,top_left_corner)   # Need this later to construct brace
    movebottom = Geom::Vector3d.new(-delta,0,0)
    trans = Geom::Transformation.new(movebottom)
    lamella.entities.transform_entities(trans,bottom_right_edge)
    
    top_and_bottom = lamella.entities.select{|e| e.is_a?(Sketchup::Face) && (e.normal == [0,0,1] or e.normal == [0,0,-1])}
    v = []
    top_and_bottom.each{|f| v << f.vertices}
    
  # Write lamella specifications

      print "\nLamella blank length "
      printf("%0.02f",lamella_blank_length)
      puts
      puts "\nLamella blank upper and lower corners"
      vf = v.flatten
      vf.sort{|a,b| a.position.z <=> b.position.z}
      vf.each{|vv| printf("\n%0.02f,%0.02f,%0.02f",vv.position.x.to_mm,vv.position.y.to_mm,vv.position.z.to_mm)}
      
      
  # Write the compound mitre saw angles (2 options, depending on orientation of lamella blank)
    
      faces = bottom_right_edge.faces
      mitre_angle = faces[0].normal.angle_between(faces[1].normal)
      puts "\nCompound mitre saw cuts (Option A)"
      print "Horizontal angle "
      printf("%0.02f",hba.radians)
      print "\nVertical angle "
      printf("%0.02f", mitre_angle.radians)
      puts
      
      edge = lamella.entities.select{|e| e != nil && e.is_a?(Sketchup::Edge) && !(e.line[1].z == 0)}[0]  # Chooses the first non-horizontal edge
      faces = edge.faces
      mitre_angle = faces[0].normal.angle_between(faces[1].normal)
      puts "\nCompound mitre saw cuts (Option B)"
      print "Horizontal angle "
      printf("%0.02f",vba.radians)
      print "\nVertical angle "
      printf("%0.02f",mitre_angle.radians)
      puts
      
  # Remove superfluous edges used in construction

    mid_edges = []
    lamella.entities.each{|e|
      if e.is_a?(Sketchup::Edge)
        mid_edges << e if (e.start.position.z == 0 and e.end.position.z == 0)
      end
          }
    lamella.entities.erase_entities(mid_edges)

  # Construct the upper curve of the lamella

    center = Geom::Point3d.new(0,-100,-roof_radius*Math.cos(theta))
    normal = Geom::Vector3d.new(0,1,0)
    radius = roof_radius*Math.cos(theta) + top_height - 0.125 # Reduce radius by 1/8" to ensure continuous upper curve
    numsides = 400
    circle = ents.add_group()
    ngon = circle.entities.add_ngon(center,normal,radius,numsides)
    face = circle.entities.add_face(ngon)
    rotate = Geom::Transformation.new([0,0,0],[0,0,1],skew_angle)
    ents.transform_entities(rotate,circle)

    face.pushpull(200,false)
    dummy = Geom::Transformation.new()
    ents.intersect_with(true,dummy,lamella,dummy,false,circle)  


  # The next section of code finds the vertices in the upper front edge of the lamella, and connects them with vertices on the upper back edge.
  # There is some complexity here because you also have to worry about cases where the curved surface intersects the bevelled ends

  # First, get an unduplicated list of the vertices above the base plane

    lents = lamella.entities
    vertices = []

    lents.each{|e|
       if e != nil && e.class == Sketchup::Edge 
         vertices << e.start if e.start.position.z > 0 
         vertices << e.end  if e.start.position.z > 0
       end
    }
     vertices.uniq!

  # Then get the vertices in front of the back plane

    fcv = []
    vertices.each{|v| fcv << v if !v.deleted? && (v.position.y == -thickness/2.0 || (v.position.x > 0 && v.position.y < thickness/2.0))}
    fcv.sort!{|a,b| a.position.x <=> b.position.x}

  # Then find the matching vertex for each, i.e. the vertices at the same height, at skew_angle from the vertex at the front

    edge_pairs = []
    fcv.each{|v|
      @fx = v.position.x
      @fy = v.position.y
      vertices.each{|a| 
        ax = a.position.x
        if ax < @fx 
          ay = a.position.y
          delta_y = (ay - @fy ) * Math.tan(skew_angle)
          delta_x = (@fx-ax)
          diff = (delta_y - delta_x).abs
          edge_pairs << [v,a]  if diff < 0.001   
        end
      }
     }

  # Finally draw the edges (there may have been a way to retain these when the intersection was made but I couldn't figure it out)
  
    for n in 0..(edge_pairs.length-1) 
      line_start = edge_pairs[n][0]
      line_end = edge_pairs[n][1]
      line = lents.add_line(line_start,line_end)
      line.find_faces
      line.smooth = "true"
      line.soft = "true"
    end

    lamella.entities.each{|e| e.material = "BurlyWood" if !e.deleted? and e.is_a?(Sketchup::Face)}
    
  #  Remove edges above the curves in lamella

    edges_above_curves = []
    lamella.entities.each{|e| 
      if e.is_a?(Sketchup::Edge) 
        edges_above_curves << e if (e.start.position.z == width/2 || e.end.position.z == top_height)
      end
        }
    lamella.entities.erase_entities(edges_above_curves) 
    
  #  Write the location of the vertices in the face of the lamella
  
    puts "\nLocations of vertices on face of lamella"
    vertices = []
    lamella.entities.each{|e| 
      if !e.deleted? && e.is_a?(Sketchup::Edge) && e.start.position.y == thickness/2 && e.end.position.y == thickness/2 
        vertices << e.start
        vertices << e.end
      end
    }
    vertices.uniq!
    vertices.each{|v| printf("%0.02f, %0.02f, %0.02f \n",v.position.x.to_mm,v.position.y.to_mm,v.position.z.to_mm)}
    puts
    puts "Exact description of elliptical upper edge"
    roof_center_x = ht * Math.tan(skew_angle)
    roof_center_y = - roof_radius*Math.cos(theta)
    printf("Center of roof in plane of front of lamella: x = %0.02f, y = %0.02f\n",roof_center_x.to_mm,roof_center_y.to_mm)
    ellipse_a = radius / Math.cos(skew_angle)
    ellipse_b = radius
    printf("Ellipse axes: major = %0.02f, minor = %0.02f \n",ellipse_a.to_mm, ellipse_b.to_mm)
    
  # Create lamella component

      lamella = lamella.to_component
      lamella.definition.name = "Lamella"
      ents.transform_entities([0,-100,0],lamella) # Push it out of the way  

# Make the reduced lamellas: half-lamellas at sills, gable lamellas at gables

  # Set up lamella groups

   left_lamellas = ents.add_group()
   left_lamellas.name = "Left lamellas"
   right_lamellas = ents.add_group()         # Right whole lamellas
   right_lamellas.name = "Right lamellas"    
   
   point = Geom::Point3d.new(0,-200,0)
   dummy = Geom::Transformation.new(point)
   gable_lamella = ents.add_instance(lamella.definition,dummy)
   gable_lamella.make_unique                 # Make gable lamella a component
   gable_lamellas = ents.add_group()         # Right lamellas at gables
   gable_lamellas.name = "Gable lamellas" 
   
   point = Geom::Point3d.new(0,-300,0)
   dummy = Geom::Transformation.new(point)
   half_lamella = ents.add_instance(lamella.definition,dummy)
   half_lamella.make_unique                  # Make half-lamella a component
   half_lamellas = ents.add_group()          # Right lamellas at sills
   half_lamellas.name = "Half lamellas"

  # Set up a transverse plane in the middle of the lamella for later measurement purposes

      pt0 = Geom::Point3d.new(0,thickness/2.0,0)
      pt1 = Geom::Point3d.new(0,-thickness/2.0,0)
      pt2 = Geom::Point3d.new(0,-thickness/2.0,3)  # Height of plane is arbitrary
      pt3 = Geom::Point3d.new(0,thickness/2.0,3)
      mid_plane = ents.add_face(pt0,pt1,pt2,pt3)
 
  # Find where lamellas intersect by setting up intersecting planes at the correct angles and position

     lents = lamella.definition.entities
     front = lents.select{|e| e.is_a?(Sketchup::Face) && e.normal == [0,1,0]}[0]
     back = lents.select{|e| e.is_a?(Sketchup::Face) && e.normal == [0,-1,0]}[0]

     fv = front.vertices
     fe = front.edges
     bv = back.vertices
     be = back.edges

     lamella_faces_dummy = ents.add_group()
     lf_ents = lamella_faces_dummy.entities
     left_front = lf_ents.add_face(fv)
     left_back = lf_ents.add_face(bv)
     roof_axis = -roof_radius * Math.cos(theta)
     rotate_vert = Geom::Transformation.new [0,0,roof_axis],[0,1,0],theta
     rotate_left = Geom::Transformation.new [0,0,0],[0,0,1], skew_angle 
     reflect = Geom::Transformation.scaling(1,-1,1)
     left_trans = rotate_vert * rotate_left * reflect
     ents.transform_entities(left_trans,left_front)
     ents.transform_entities(left_trans,left_back)
     ents.transform_entities(left_trans,mid_plane)  
     inter_node_distance = 2.0 * roof_radius * Math.sin(theta) * Math.tan(skew_angle)
     inter_node_distance = inter_node_distance + (thickness*(1+1/Math.cos(2*skew_angle))) / Math.cos(skew_angle) + bolt_diameter * Math.tan(2*skew_angle)/Math.cos(skew_angle)
     half_ind = inter_node_distance / 2.0
     move_right = [0,-half_ind,0]
     move_left = [0,half_ind,0]
     rotate_left = Geom::Transformation.new [0,-half_ind,0],[0,0,1], skew_angle
     left_trans = rotate_left
     ents.transform_entities(left_trans,left_front)
     ents.transform_entities(left_trans,left_back)
     ents.transform_entities(left_trans,mid_plane)

     ents.transform_entities(move_left,left_front)
     ents.transform_entities(move_left,left_back)
     ents.transform_entities(move_left,mid_plane)

     front_line = Geom.intersect_plane_plane(left_back.plane,front.plane)
     pts = []
     fe.each{|e| 
       pt = Geom.intersect_line_line(front_line,e.line)
       pts << pt if (pt != nil and pt.z >= -width/2)
       }  # Find where line intersects edges
     pts.sort!{|a,b| a.z <=> b.z}
     pt0 = pts[0]
     pt1 = pts[1]
     pt3 = Geom::Point3d.new(pt0.x+thickness*Math.tan(skew_angle),-thickness/2,-width/2)

     back_line = [pt3,front_line[1]]
     pts = []
     be.each{|e| 
       pt = Geom.intersect_line_line(back_line,e.line)
       pts << pt if (pt != nil and pt.z >= -width/2)
       }  # Find where line intersects edges
     pts.sort!{|a,b| a.z <=> b.z}
     pt2 = pts[1]
     
     left_end_face = ents.add_face(pt0,pt1,pt2,pt3)  # We'll need this copy of the end of the half-lamella later to figure out where the beams go

     hents = half_lamella.definition.entities
     end_face = hents.add_face(pt0,pt1,pt2,pt3)
     angle_of_end = end_face.normal.angle_between(front.normal)  # Need to figure out angle between end face and face of lamella for later

     end_face_edges = end_face.edges

     all_edges = []
     hents.each{|h|
       all_edges << h if !h.deleted? and h.is_a?(Sketchup::Edge)
     }

     edges_less_end_face_edges = all_edges - end_face_edges
     edges_to_erase = []
     edges_less_end_face_edges.each{ |h|
           if h.end.position.z == -width/2
             edges_to_erase << h  if h.start.position.y > 0 and (h.start.position.x < pt0.x) || (h.end.position.x < pt0.x)
             edges_to_erase << h  if h.start.position.y < 0 and (h.start.position.x < pt3.x) || (h.end.position.x < pt3.x)
           else
             edges_to_erase << h if h.end.position.y > 0 and (h.start.position.x < pt1.x) || (h.end.position.x < pt1.x)
             edges_to_erase << h if h.end.position.y < 0 and (h.start.position.x < pt2.x) || (h.end.position.x < pt2.x)
           end
       }
   # Chop off the extra half

     hents.erase_entities(edges_to_erase)

   # Write the half-lamella specifications

     puts "\nEnd face of the half-lamella"
     end_face.vertices.each{|v| printf("%0.02f, %0.02f, %0.02f \n",v.position.x.to_mm,v.position.y.to_mm,v.position.z.to_mm)}
     puts "Compound mitre saw cuts"
     inclined_edge = end_face.edges.select{|e| e.start.position.z == - hw && e.end.position.z > 0}[0]
     ip = inclined_edge.start.position
     pt2 = Geom::Point3d.new([ip.x + 10,ip.y, ip.z])
     cline = ents.add_cline(ip,pt2)
     angle_between = inclined_edge.line[1].angle_between(cline.direction)
     printf("Horizontal angle %0.02f\n",180 - angle_of_end.radians)
     printf("Vertical angle %0.02f\n",angle_between.radians )
     
   # Now figure out the gable lamella dimensions
   
     back_line = Geom.intersect_plane_plane(left_front.plane,back.plane)

     pts = []
     be.each{|e|
       pt = Geom.intersect_line_line(back_line,e.line)
       pts << pt if (pt != nil and pt.z >= -hw)
       }  # Find where line intersects back edges
     pts.sort!{|a,b| a.z <=> b.z}
     pt0 = pts[0]
     pt1 = pts[1] 

     front_line = Geom.intersect_plane_plane(left_front.plane,front.plane)
     pts = []
     fe.each{|e| 
       pt = Geom.intersect_line_line(front_line,e.line)  # Find where line intersects front edges
       pts << pt if (pt != nil and pt.z >= -hw)
     }
     pts.sort!{|a,b| a.z <=> b.z}
     pt3 = pts[0]
     pt2 = pts[1]

     gents = gable_lamella.definition.entities
     gable_end_face = gents.add_face(pt0,pt1,pt2,pt3)

     ents_to_erase = []
       gents.each{|h|
         if h.deleted?
         else
           if h.is_a?(Sketchup::Edge)
             ents_to_erase << h if (h.start.position.x > pt2.x) || (h.end.position.x > pt2.x)

        end
        end
         }
       gents.erase_entities(ents_to_erase)


   # Write the gable lamella specifications

       puts "\nEnd face of the gable lamella"
       gable_end_face.vertices.each{|v| printf("%0.02f, %0.02f, %0.02f \n",v.position.x.to_mm,v.position.y.to_mm,v.position.z.to_mm)}
       puts
# Now we draw the arches

  # Set up transformations

    rotate_left = Geom::Transformation.new([0,0,0],[0,0,1], skew_angle)
    rotate_right = Geom::Transformation.new([0,0,0],[0,0,1], -skew_angle)

if lamellas % 2 == 0     #  Number of lamellas is even
  
  # Add whole left lamellas
    n = 1
    while n <= lamellas / 2
      angle = (2*n-1)*theta
      rotate_left_vert = Geom::Transformation.new [0,0,roof_axis],[0,1,0],angle
      rotate_right_vert = Geom::Transformation.new [0,0,roof_axis],[0,1,0],-angle
      left_lamella_instance = left_lamellas.entities.add_instance(lamella.definition,rotate_left_vert * rotate_left * reflect)
      left_lamella_instance.name = "Left #{lamellas / 2 - n}"
      left_lamella_instance = left_lamellas.entities.add_instance(lamella.definition,rotate_right_vert * rotate_left * reflect)
      left_lamella_instance.name = "Left #{lamellas / 2 + n - 1}"
      n +=1
    end

    low_lamella_transformation = rotate_right_vert * rotate_left * reflect  # We'll need this later to set up the lamella braces
    
  # Add whole right lamellas and gable lamellas
    n = 1
    while n <= (lamellas/ 2)-1
      angle = (2*n)*theta
      rotate_left_vert = Geom::Transformation.new([0,0,roof_axis],[0,1,0],angle)
      rotate_right_vert = Geom::Transformation.new([0,0,roof_axis],[0,1,0],-angle)
      right_lamella_instance = right_lamellas.entities.add_instance(lamella.definition,rotate_left_vert * rotate_right)
      right_lamella_instance.name = "Right #{lamellas / 2 - n}"
      new_gable_lamella = gable_lamellas.entities.add_instance(gable_lamella.definition,rotate_left_vert * rotate_right)
      new_gable_lamella.name = "Gable #{lamellas / 2 - n}"
      right_lamella_instance = right_lamellas.entities.add_instance(lamella.definition,rotate_right_vert * rotate_right)
      right_lamella_instance.name = "Right #{lamellas / 2 + n}"
      new_gable_lamella = gable_lamellas.entities.add_instance(gable_lamella.definition,rotate_right_vert * rotate_right)
      new_gable_lamella.name = "Gable #{lamellas / 2 + n}"
      n +=1
    end
  
  # Add single horizontal right and gable lamellas
    top_right_lamella = right_lamellas.entities.add_instance(lamella.definition,rotate_right)
    top_right_lamella.name = "Top right lamella"
    new_gable_lamella = gable_lamellas.entities.add_instance(gable_lamella.definition,rotate_right)
    new_gable_lamella.name = "Top gable lamella"
  
  # Add gable lamella at sill
    rotate_right_vert = Geom::Transformation.new([0,0,roof_axis],[0,1,0],lamellas * theta)
    new_gable_lamella = gable_lamellas.entities.add_instance(gable_lamella.definition,rotate_right_vert * rotate_right)
    new_gable_lamella.name = "Gable #{lamellas / 2 + n}"

  # Add half lamellas  at sill
    angle = lamellas * theta
    trans = Geom::Transformation.new([0,0,roof_axis],[0,1,0],-angle)
    half_lamella = half_lamellas.entities.add_instance(half_lamella.definition,trans * rotate_right)
    half_lamella.name = "Half-lamella +"
  
    half_lamella.definition.name = "Half lamella"
    swing = Geom::Transformation.new([0,0,0],[0,0,1],180.degrees)
    half_lamella_instance = half_lamellas.entities.add_instance(half_lamella.definition,swing*trans * rotate_right)
    half_lamella_instance.name = "Half-lamella -"

else  # Number of lamellas is odd
  
  # Add whole left lamellas
    n = 1
    while n <= (lamellas - 1) / 2
      angle = 2*n*theta
      rotate_left_vert = Geom::Transformation.new [0,0,roof_axis],[0,1,0],angle
      rotate_right_vert = Geom::Transformation.new [0,0,roof_axis],[0,1,0],-angle
      left_lamella_instance = left_lamellas.entities.add_instance(lamella.definition,rotate_left_vert * rotate_left * reflect)
      left_lamella_instance.name = "Left #{lamellas / 2 - n}"
      left_lamella_instance = left_lamellas.entities.add_instance(lamella.definition,rotate_right_vert * rotate_left * reflect)
      left_lamella_instance.name = "Left #{lamellas / 2 + n - 1}"
      n +=1
    end

    low_lamella_transformation = rotate_right_vert * rotate_left * reflect  # We'll need this later to set up the lamella braces
    
  # Add single horizontal left lamella
  
    top_left_lamella = left_lamellas.entities.add_instance(lamella.definition,rotate_left * reflect)
    top_left_lamella.name = "Top right lamella"
  
  # Add whole right lamellas and gable lamellas
    n = 1
    while n <= (lamellas-1) / 2
      angle = (2*n - 1)*theta
      rotate_left_vert = Geom::Transformation.new([0,0,roof_axis],[0,1,0],angle)
      rotate_right_vert = Geom::Transformation.new([0,0,roof_axis],[0,1,0],-angle)
      right_lamella_instance = right_lamellas.entities.add_instance(lamella.definition,rotate_left_vert * rotate_right)
      right_lamella_instance.name = "Right #{lamellas / 2 - n}"
      new_gable_lamella = gable_lamellas.entities.add_instance(gable_lamella.definition,rotate_left_vert * rotate_right)
      new_gable_lamella.name = "Gable #{lamellas / 2 - n}"
      right_lamella_instance = right_lamellas.entities.add_instance(lamella.definition,rotate_right_vert * rotate_right)
      right_lamella_instance.name = "Right #{lamellas / 2 + n}"
      new_gable_lamella = gable_lamellas.entities.add_instance(gable_lamella.definition,rotate_right_vert * rotate_right)
      new_gable_lamella.name = "Gable #{lamellas / 2 + n}"
      n +=1
    end
  
  # Add gable lamella at sill
    rotate_right_vert = Geom::Transformation.new([0,0,roof_axis],[0,1,0],lamellas * theta)
    new_gable_lamella = gable_lamellas.entities.add_instance(gable_lamella.definition,rotate_right_vert * rotate_right)
    new_gable_lamella.name = "Gable #{lamellas / 2 + n}"

  # Add half lamellas  at sill
    angle = lamellas * theta
    trans = Geom::Transformation.new([0,0,roof_axis],[0,1,0],-angle)
    half_lamella_instance = half_lamellas.entities.add_instance(half_lamella.definition,trans * rotate_right)
    half_lamella_instance.name = "Half-lamella +"
  
    half_lamella.definition.name = "Half lamella"
    swing = Geom::Transformation.new([0,0,0],[0,0,1],180.degrees)
    half_lamella_instance = half_lamellas.entities.add_instance(half_lamella.definition,swing*trans * rotate_right)
    half_lamella_instance.name = "Half-lamella -" 
  
end
 
  # Find the half-lamellas and move them into position
    hl = half_lamellas.entities.find{|e| e.name == "Half-lamella +"}
    move_left = [0,half_ind,0]
    half_lamellas.entities.transform_entities(move_left,hl)
   
    hr = half_lamellas.entities.find{|e| e.name == "Half-lamella -"}
    move_right = [0,-half_ind,0]
    half_lamellas.entities.transform_entities(move_left,hr)
     
  # Move the right lamellas into position 
    ents.transform_entities(move_left,right_lamellas)

  # Set up the gable lamellas at the other end of the roof
    ents.transform_entities(move_right,gable_lamellas) 
    other_end_gable_lamellas = gable_lamellas.copy
    ents.transform_entities(swing,other_end_gable_lamellas)
    move_length_of_roof = [0,((number_of_arches) * inter_node_distance),0]
    ents.transform_entities(move_length_of_roof,other_end_gable_lamellas)

  # Set up the repeating roof elements

  # Define a method to replicate a component: borrowed from Martin Rinehart's Sketchtalk

      def mc(component,distance,number_of_copies)
        ents = Sketchup.active_model.entities
        defi = component.definition
        trans = component.transformation
        pt = trans.origin

        for i in (1..number_of_copies)
          pt += distance
          trans = Geom::Transformation.new(pt)
          ents.add_instance(defi,trans)
        end
      end

  # Turn the groups into components to make it easier to replicate new instances
      left_lamellas = left_lamellas.to_component
      right_lamellas = right_lamellas.to_component
      half_lamellas = half_lamellas.to_component

  # Move/copy the several elements
      inter_node_distance_vector = Geom::Vector3d.new(0,inter_node_distance,0)
      mc(left_lamellas,inter_node_distance_vector,number_of_arches)
      mc(right_lamellas,inter_node_distance_vector,number_of_arches-1)
      mc(half_lamellas,inter_node_distance_vector,number_of_arches-1)
 
  # Set up beams at the edges of the roof
    
    ents.transform_entities(trans * rotate_right, left_end_face)
    ents.transform_entities(move_left,left_end_face)
    normal = left_end_face.normal
    beam_angle = Math::atan(normal.z / normal.x)
    print "\nBeam angle "
    printf("%0.02f",beam_angle.radians)
    puts
    
    low_points = left_end_face.vertices.sort{|a,b| a.position.y <=> b.position.y}

    beam = ents.add_group()
    
    edge_of_left_lamella = ents.add_line(slanted_edge.start.position,slanted_edge.end.position)
    ents.transform_entities(low_lamella_transformation,edge_of_left_lamella)
    end_face_y = edge_of_left_lamella.end.position.y
    pts =[]
    pts[0] = [low_points[3].position.x,end_face_y,low_points[3].position.z]
    pts[1] = [low_points[2].position.x,end_face_y,low_points[2].position.z]
    pts[2] = [low_points[2].position.x,end_face_y,low_points[3].position.z]
    beam_end = beam.entities.add_face pts
    beam_end.material = "Chocolate"
    beam_end.reverse!
    beam_end.pushpull( ((number_of_arches+1)*inter_node_distance - thickness/Math.cos(skew_angle)),true)
    beam = beam.to_component
    reflect_x = Geom::Transformation.scaling(-1,1,1)
    other_beam = ents.add_instance(beam.definition,reflect_x)

 #  Make the bolt, washers, and nuts  
    
    bolt_length = 2.0 * thickness + thickness / Math.cos(2*skew_angle) + bolt_diameter * Math.tan(2*skew_angle)
    half_bolt = lents.add_group()

    centre = Geom::Point3d.new(bolt_length/2.0 - 10.mm,0,0)
    normal = Geom::Vector3d.new(-1,0,0)
    bolt_end = half_bolt.entities.add_circle(centre,normal,bolt_diameter/2.0)
    face = half_bolt.entities.add_face(bolt_end)
    face.pushpull((bolt_length-20.mm)/2)

    centre = Geom::Point3d.new(bolt_length/2.0,0,0)
    normal = Geom::Vector3d.new(-1,0,0)
    washer = half_bolt.entities.add_circle(centre,normal,0.25 + bolt_diameter/2.0)
    face = half_bolt.entities.add_face(washer)  # washer radius is 1/4" greater than bolt
    face.pushpull(10.mm)  # washer will sink into lamella

    centre = Geom::Point3d.new(bolt_length/2.0+6.mm,0,0)
    normal = Geom::Vector3d.new(-1,0,0)
    nut = half_bolt.entities.add_ngon(centre,normal,bolt_diameter/2.0,6)
    face = half_bolt.entities.add_face(nut)
    face.pushpull(6.mm)  # 1/4" thick nut
    half_bolt = half_bolt.to_component

    bolt = lents.add_group()
    dummy = Geom::Transformation.new()
    bolt.entities.add_instance(half_bolt.definition,dummy) # Add the half-bolt in place
    rotate_left = Geom::Transformation.new([0,0,0],[0,0,1], -Math::PI)
    bolt.entities.add_instance(half_bolt.definition,rotate_left)  #  Add a half-bolt rotated 180 degrees
    bolt = bolt.to_component

  # Figure out where the bolts go

    left_front_front_edge = left_front.edges.select{|p| p.start.position.y == p.end.position.y}[0]
    left_back_front_edge = left_back.edges.select{|p| p.start.position.y == p.end.position.y}[0]
    front_corner = left_front_front_edge.vertices.min{|e,f| e.position.x <=> f.position.x }
    back_corner = left_back_front_edge.vertices.min{|e,f| e.position.x <=> f.position.x }
    height_of_bolt = (front_corner.position.z + back_corner.position.z)/2.0

    if number_of_bolts == 1
      rotate_left = Geom::Transformation.new([0,0,0],[0,0,1], -(Math::PI/2 - 2*skew_angle))
      where_bolt_goes = Geom::Point3d.new([0,0,height_of_bolt + hw])
      move_up = Geom::Transformation.new(where_bolt_goes) # Move it into position
      lents.add_instance(bolt.definition,move_up*rotate_left)

      bolt_axis = ents.add_line(centre,centre+normal)
      ents.transform_entities(move_up * rotate_left,bolt_axis)

      front_centre_hole = Geom.intersect_line_plane(bolt_axis.line,front.plane)
      back_centre_hole = Geom.intersect_line_plane(bolt_axis.line,back.plane)
      printf("\nCentre hole position: Front %0.02f, %0.02f, %0.02f",front_centre_hole.x.to_mm,front_centre_hole.y.to_mm,front_centre_hole.z.to_mm)
      printf("\nCentre hole position: Back %0.02f, %0.02f, %0.02f",back_centre_hole.x.to_mm,back_centre_hole.y.to_mm,back_centre_hole.z.to_mm)
      front_end_hole = Geom.intersect_line_plane(bolt_axis.line,left_front.plane)
      back_end_hole = Geom.intersect_line_plane(bolt_axis.line,left_back.plane)
      puts "End hole positions:"
      bottom_front = left_front.edges.max{|a,b| a.length <=> b.length}
      bottom_back = left_back.edges.max{|a,b| a.length <=> b.length}
      front_z = front_end_hole.distance_to_line(bottom_front.line)
      back_z = back_end_hole.distance_to_line(bottom_back.line)
      printf("Z front and back %0.02f,%0.02f \n",front_z.to_mm,back_z.to_mm)
      printf("Z angle = %0.02f \n",(Math.atan(thickness/(front_z - back_z))).radians)
      front_x = front_end_hole.distance_to_plane(mid_plane.plane)
      back_x = back_end_hole.distance_to_plane(mid_plane.plane)
      printf("X front and back %0.02f, %0.02f \n",front_x.to_mm,back_x.to_mm)
      printf("X angle = %0.02f \n",(Math.atan(thickness/(front_x - back_x))).radians)
      ents.erase_entities(bolt_axis)
    else
      
      bolt_number = 1
      while bolt_number <= 2
        puts "\nBolt #{bolt_number}"
        rotate_left = Geom::Transformation.new([0,0,0],[0,0,1], -(Math::PI/2 - 2*skew_angle))
        where_bolt_goes = Geom::Point3d.new([0,0,height_of_bolt + 2 * bolt_number * hw / 3])
        move_up = Geom::Transformation.new(where_bolt_goes) # Move it into position
        lents.add_instance(bolt.definition,move_up*rotate_left)

        bolt_axis = ents.add_line(centre,centre+normal)
        ents.transform_entities(move_up * rotate_left,bolt_axis)

        front_centre_hole = Geom.intersect_line_plane(bolt_axis.line,front.plane)
        back_centre_hole = Geom.intersect_line_plane(bolt_axis.line,back.plane)
        printf("\nCentre hole position: Front %0.02f, %0.02f, %0.02f",front_centre_hole.x.to_mm,front_centre_hole.y.to_mm,front_centre_hole.z.to_mm)
        printf("\nCentre hole position: Back %0.02f, %0.02f, %0.02f",back_centre_hole.x.to_mm,back_centre_hole.y.to_mm,back_centre_hole.z.to_mm)
        front_end_hole = Geom.intersect_line_plane(bolt_axis.line,left_front.plane)
        back_end_hole = Geom.intersect_line_plane(bolt_axis.line,left_back.plane)
        puts "\nEnd hole position:"
        bottom_front = left_front.edges.max{|a,b| a.length <=> b.length}
        bottom_back = left_back.edges.max{|a,b| a.length <=> b.length}
        front_z = front_end_hole.distance_to_line(bottom_front.line)
        back_z = back_end_hole.distance_to_line(bottom_back.line)
        printf("Z front and back %0.02f,%0.02f \n",front_z.to_mm,back_z.to_mm)
        printf("Z angle = %0.02f \n",(Math.atan(thickness/(front_z - back_z))).radians)
        front_x = front_end_hole.distance_to_plane(mid_plane.plane)
        back_x = back_end_hole.distance_to_plane(mid_plane.plane)
        printf("X front and back %0.02f, %0.02f \n",front_x.to_mm,back_x.to_mm)
        printf("X angle = %0.02f \n",(Math.atan(thickness/(front_x - back_x))).radians)
        bolt_number = bolt_number + 1
        ents.erase_entities(bolt_axis)
     end
    end

  # Build the lamella brace

    bottom_edge_of_beam = [left_end_face.vertices[2],Geom::Vector3d.new(0,1,0)]
    top_edge_of_beam = [left_end_face.vertices[3].position,Geom::Vector3d.new(0,1,0)]
    
  # The left lamella will be fractionally off the position of the right lamella ( < 0.1mm) but enough to 
  # screw up the definition of the base of the brace, so we have to project on to the plane

    ps = edge_of_left_lamella.start.position.project_to_plane(left_end_face.plane)
    pe = edge_of_left_lamella.end.position.project_to_plane(left_end_face.plane)
    projected_edge = ents.add_line(ps,pe)
    top_right_corner = Geom.intersect_line_line(projected_edge.line,top_edge_of_beam)
    bottom_right_corner = Geom.intersect_line_line(projected_edge.line,bottom_edge_of_beam)
    brace = ents.add_group()
    base_of_brace = brace.entities.add_face(top_right_corner,bottom_right_corner,left_end_face.vertices[2].position,left_end_face.vertices[3].position)
    base_of_brace.reverse!
    base_of_brace.pushpull(thickness,true)
    top_of_brace = brace.entities.select{|b| b.is_a?(Sketchup::Face) and b.normal == base_of_brace.normal and b.plane != base_of_brace.plane}
    top_left_brace_edge = brace.entities.add_line(top_of_brace[0].vertices[0].position,top_of_brace[0].vertices[1].position)
    top_right_brace_edge = brace.entities.add_line(top_of_brace[0].vertices[2].position,top_of_brace[0].vertices[3].position)
    brace_bevel = 180 - angle_of_end.radians
    adj = thickness/Math::tan(brace_bevel.degrees)
    movetop = Geom::Vector3d.new(0,adj,0)
    trans = Geom::Transformation.new(movetop)
    brace.entities.transform_entities(trans,top_left_brace_edge)
    movetop = Geom::Vector3d.new(0,-adj,0)
    trans = Geom::Transformation.new(movetop)
    brace.entities.transform_entities(trans,top_right_brace_edge)
    brace = brace.to_component
    braces = ents.add_group()
    dummy = Geom::Transformation.new()
    braces.entities.add_instance(brace.definition,dummy)
    braces.entities.add_instance(brace.definition,swing)
    braces = braces.to_component
    mc(braces,inter_node_distance_vector,number_of_arches)

  # Create a horizontal instance of a brace at the origin, to get the dimensions
  
    trc = top_right_corner
    brace_vector = Geom::Vector3d.new(-trc.x,-trc.y,-trc.z)
    move_brace = Geom::Transformation.new(brace_vector)
    brace_axis = Geom::Vector3d.new(0,1,0)
    rotate_brace = Geom::Transformation.new([0,0,0],brace_axis,-(90.degrees - beam_angle))
    brace_transformation = Geom::Transformation.new(rotate_brace*move_brace)
    brace_template = ents.add_instance(brace.definition,brace_transformation)
    bte = brace_template.explode
    vertices = bte.select{|e| e.is_a?(Sketchup::Vertex)}
  
  # Write the brace specifications
  
    puts "\nBrace specifications (coordinates of corners)"
    vertices.each{|v| printf("%0.02f, %0.02f, %0.02f \n",v.position.x.to_mm,v.position.y.to_mm,v.position.z.to_mm)}
    
  # Erase the temporary brace
  
    bte.each{|e| e.erase! if e.valid? && e.is_a?(Sketchup::Drawingelement)}
    print "\nBrace bevel "
    printf("%0.02f",brace_bevel)
    puts
    
  # Write out the gable section data
  
    inner = left_end_face.vertices[3]
    inner_radius = ( (inner.position.x)**2 + (inner.position.z - roof_center_y)**2 ) ** 0.5
    printf("Section at gable end: inner radius = %0.01f; outer radius = %0.01f\n",inner_radius.to_mm,radius.to_mm)
    
  
  # A kludge to line up the beam with the braces (just to be pretty)

    off_centre_delta = (left_end_face.vertices[2].position.y).abs - (bottom_right_corner.y).abs
    move_beam = Geom::Vector3d.new(0,-off_centre_delta,0)
    trans = Geom::Transformation.new(move_beam)
    ents.transform_entities(trans,other_beam)

  # Remove construction elements

  ents.erase_entities(circle)
  ents.erase_entities(left_front.edges)
  ents.erase_entities(left_back.edges)
  ents.erase_entities(mid_plane.edges)
  #ents.erase_entities(half_lamella)
  #ents.erase_entities(gable_lamella)

  #lamella.erase!
  
  half_bolt.erase!
  left_end_face.edges.each{|e| e.erase!}

# Put the plug-in in the menu if it isn't there already

  if( not file_loaded?("lamella_roof.rb") )

      plugins = UI.menu("Draw")
      plugins.add_separator
      plugins.add_item("saghf ba choob char tarash") { lamella_roof }

  end

  file_loaded("lamella_roof.rb")

end
